/* Flex declarations */

%option noyywrap
%option debug
%option caseless
%option 8bit
%option never-interactive

%x STATE_DEFCMD
%x STATE_COMMENT
%x STATE_INCLUDE
%x STATE_DATA

%{
/* C declarations */
#include <ctype.h>

#include "qb.tab.h"

typedef struct saved_yy_context
{
	int line;
	int col;
	struct saved_yy_context *next;
	YY_BUFFER_STATE state;
	FILE *fp;
} saved_yy_context;
static saved_yy_context *contexts = NULL;
static saved_yy_context current_context = { 1, 0, NULL, NULL, NULL };
#define YY_INPUT(buf,result,max_size) \
  { \
    int c; do c = fgetc(yyin); while(c == '\r'); \
    if(c == EOF) result=YY_NULL; \
    else { \
      if(c == '\n')      { ++current_context.line; current_context.col=0; } \
      else if(c == '\t') { current_context.col=(current_context.col+7)&~7; } \
      else { ++current_context.col; } \
      buf[0]=c; result=1; \
    } \
  }

void yyerror(const char *msg);
void yyerror(const char *msg)
{
    fprintf(stderr,
        "\n\n"
        "At line %d column %d\n"
        "\t%s\n",
        current_context.line,
        current_context.col,
        msg
    );
}

static int cur_deftype;

%}
/* Flex declarations */

digit [0-9]
blank [ \t]

%%
	/* Grammar begins: */

	/* Exception: DEFTYPE commands */
DEFINT     { BEGIN(STATE_DEFCMD); cur_deftype = '%'; }
DEFSNG     { BEGIN(STATE_DEFCMD); cur_deftype = '!'; }
DEFDBL     { BEGIN(STATE_DEFCMD); cur_deftype = '#'; }
DEFSTR     { BEGIN(STATE_DEFCMD); cur_deftype = '$'; }
DEFLNG     { BEGIN(STATE_DEFCMD); cur_deftype = '&'; }

<STATE_DEFCMD>[A-Z]-[A-Z] {
    int begin = toupper(yytext[0]), end = toupper(yytext[2]);
    fprintf(stderr, "DEF-%c character range %c-%c\n", cur_deftype, begin, end);
  }
<STATE_DEFCMD>[A-Z] {
    int pick = toupper(yytext[0]);
    fprintf(stderr, "DEF-%c character %c\n", cur_deftype, pick);
  }
<STATE_DEFCMD>[, \t]  { /* ignore comma and space */ }
<STATE_DEFCMD>:       { BEGIN(INITIAL); return COLONCHAR; /* end of def */ }
<STATE_DEFCMD>\n      { BEGIN(INITIAL); return CRLF;      /* end of def */ }
<STATE_DEFCMD>\'      { BEGIN(STATE_COMMENT); }
<STATE_DEFCMD>.       { BEGIN(INITIAL); return UNKNOWN_CHARACTER; }


        /* Exception: DATA */
DATA       { BEGIN(STATE_DATA); return DATA; }
<STATE_DATA>\"[^"\n]*["\n]  {
    fprintf(stderr, "Data string (quoted): '%s'\n", yytext);
    return STRING;
  }
<STATE_DATA>[ \t] { /* Ignore space in DATA */ }
<STATE_DATA>[^:,\n]+  {
    fprintf(stderr, "Data string (unquoted): '%s'\n", yytext);
    /* FIXME: ltrim and rtrim the string - qb does that too. */
    return STRING;
  }
<STATE_DATA>,    { return *yytext; }
<STATE_DATA>:    { BEGIN(INITIAL); return COLONCHAR; /* end of data */ }
<STATE_DATA>\n   { BEGIN(INITIAL); return CRLF;      /* end of data */ }



	/* Exception: Comments */
REM\       { BEGIN(STATE_COMMENT); }
\'         { BEGIN(STATE_COMMENT); }

<STATE_COMMENT>\$DYNAMIC {
    fprintf(stderr, "Meta $DYNAMIC\n");
    /* Don't return anything */
  }
<STATE_COMMENT>\$STATIC {
    fprintf(stderr, "Meta $STATIC\n");
    /* Don't return anything */
  }
<STATE_COMMENT>\$INCLUDE:\ *\' { BEGIN(STATE_INCLUDE); }
<STATE_COMMENT>\n        { BEGIN(INITIAL); return CRLF; }
<STATE_COMMENT>[^$\n]+   { /* Ignore until $ */ }
<STATE_COMMENT>.         { /* Ignore other   */ }

<STATE_INCLUDE>[^']* {
    fprintf(stderr, "Meta $INCLUDE: '%s'\n", yytext);
    {
      current_context.state = YY_CURRENT_BUFFER;
      current_context.fp    = yyin;
      current_context.next  = contexts;
      contexts = (saved_yy_context *)malloc(sizeof(*contexts));
      *contexts = current_context;
      
      yyin = fopen(yytext, "rt");
      if(!yyin) { YY_FATAL_ERROR("unable to open include file"); }
      /*setbuf(yyin, NULL);*/
      
      yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
      current_context.line = 1;
      current_context.col  = 0;
      BEGIN(INITIAL);
    }
  }
<STATE_INCLUDE>\'        { BEGIN(STATE_COMMENT); }

	/* Logical operators */
IMP        return IMP;
EQV        return EQV;
XOR        return XOR;
OR         return  OR;
AND        return AND;
NOT        return NOT;

MOD        return MOD;

	/* If stuff */
IF         return IF;
THEN       return THEN;
ELSE       return ELSE;
ELSEIF     return ELSEIF;
"END IF"   return ENDIF;

	/* Switch stuff */
SELECT     return SELECT;
CASE       return CASE;
"END SELECT"  return ENDSELECT;
IS         return IS;

	/* Loop stuff */
"DO WHILE" return DOWHILE;
"DO UNTIL" return DOUNTIL;
"LOOP WHILE" return LOOPWHILE;
"LOOP UNTIL" return LOOPUNTIL;
WEND       return WEND;
DO         return DO;
LOOP       return LOOP;
WHILE      return WHILE;
"EXIT DO"  return EXITDO;

	/* Sub stuff */
SUB           return SUB;
"END SUB"     return ENDSUB;
"EXIT SUB"    return EXITSUB;
"DECLARE SUB" return DECLARESUB;
FUNCTION           return FUNCTION;
"END FUNCTION"     return ENDFUNCTION;
"EXIT FUNCTION"    return EXITFUNCTION;
"DECLARE FUNCTION" return DECLAREFUNCTION;

	/* For stuff */
FOR        return FOR;
TO         return TO;
STEP       return STEP;
NEXT       return NEXT;
"EXIT FOR" return EXITFOR;

	/* Typedef stuff */
INTEGER    return INTEGER;
SINGLE     return SINGLE;
DOUBLE     return DOUBLE;
STRING     return STRING;
LONG       return LONG;
AS         return AS;
TYPE       return TYPE;
"END TYPE" return ENDTYPE;
SHARED     return SHARED;
DIM        return DIM;
REDIM      return REDIM;
CONST      return CONST;

	/* Other */
GOTO       return GOTO;
GOSUB      return GOSUB;
RETURN     return RETURN;
RESTORE    return RESTORE;
CALL       return CALL;

END        return END;
LET        return LET;

PRINT      return PRINT;
USING      return USING;
LPRINT     return LPRINT;

ON         return ON;
OFF        return OFF;
PLAY       return PLAY;
STOP       return STOP;
POKE       return POKE;
SYSTEM     return SYSTEM;
LOCATE     return LOCATE;
ERROR      return ERROR;
RESUME     return RESUME;
READ       return READ;
OPEN       return OPEN;

OUTPUT     return OUTPUT;
INPUT      return INPUT;
APPEND     return APPEND;
RANDOM     return RANDOM;
BINARY     return BINARY;
WRITE      return WRITE;
ACCESS     return ACCESS;
LOCK       return LOCK;
CLOSE      return CLOSE;

INKEY\$    return INKEY_S;
CSRLIN     return CSRLIN;
FREEFILE   return FREEFILE;
SPACE\$    return SPACE_S;
POS        return POS;
LEN        return LEN;

"DEF SEG"  return DEFSEG;

[a-z][a-z0-9]*[!#%$&]? {
    fprintf(stderr, "Identifier: '%s' (%u)\n", yytext, yyleng);
    return IDENTIFIER;
  }

-?{digit}+[%&] {
    fprintf(stderr, "Integer: '%s'\n", yytext);
    return NUMBER;
  }

&H[0-9A-F]+[%&]? {
    fprintf(stderr, "Hex integer: '%s'\n", yytext);
    return NUMBER;
  }

-?{digit}+(\.{digit}*)?(e[+-]?{digit}+)?[!#]? {
    fprintf(stderr, "Numeric: '%s'\n", yytext);
    return NUMBER;
  }

#{digit}* {
    fprintf(stderr, "File number is: '%s'\n", yytext);
    /* FIXME: ignore first character */
    return FILENUM;
  }

:       { BEGIN(INITIAL); return COLONCHAR; }
\n      { BEGIN(INITIAL); return CRLF; }

"<="    return LESSOREQUAL;
">="    return GREATEROREQUAL;
"<>"    return NOTEQUAL;
"><"    return NOTEQUAL;
"!="    return NOTEQUAL;

\"[^"\n]*\"  {
    fprintf(stderr, "String: '%s'\n", yytext);
    return STRING;
  }

[-<>=+*/\\,;().^] { return *yytext; }

{blank}+    /* blank   */

<<EOF>> {
    fclose(yyin);
    if(contexts)
    {
      current_context = *contexts;
      yyin = current_context.fp;
      free(contexts);
      contexts = current_context.next;
      
      yy_delete_buffer(YY_CURRENT_BUFFER);
      yy_switch_to_buffer(current_context.state);
    }
    else
    {
      yyterminate();
    }
  }

. { BEGIN(INITIAL); return UNKNOWN_CHARACTER; }

%%
